package com.gct.dlview.thrift;

import com.gct.dlview.thrift.train.TrainServerWrapper;
import lombok.extern.slf4j.Slf4j;
import org.apache.thrift.TException;
import org.apache.thrift.protocol.TBinaryProtocol;
import org.apache.thrift.transport.TSocket;
import org.apache.thrift.transport.TTransport;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.stream.Collectors;

@Component
@Slf4j
public class HealthyManager implements IHealthyManage {


    @Autowired
    private ConfigProvider configProvider;

    @Override
    public void checkHealthy(Serializable serverId) {
        ServerConfig config = configProvider.getValue(serverId);
        if (config == null) {
            log.info("Server is not available ,specified id={},next time check healthy state.", serverId);
            return;
        }
        try (TrainServerWrapper server = new TrainServerWrapper(config)) {
            config.setRunState(server.healthyCheck());
        } catch (Exception e) {
            log.error("Train server can not connect,server id:{},name:{}", config.getId(), config.getName());
            config.setRunState(ServerRunState.MISSING.getCode());
        }
//        TTransport transport = new TSocket(config.getHost(), config.getPort(), config.getTimout());
//        com.gct.dlview.thrift.train.TrainServer.Client client = new com.gct.dlview.thrift.train.TrainServer.Client(new TBinaryProtocol(transport));
//        try {
//            transport.open();
//        } catch (TException e) {
//            config.setRunState(ServerRunState.MISSING.getCode());
//        }
//
//        try {
//            if (client.healthyCheck())
//                config.setRunState(ServerRunState.READY.getCode());
//        } catch (TException e) {
//            config.setRunState(ServerRunState.BUSY.getCode());
//        }
//        if (transport.isOpen()) transport.close();

        configProvider.updateCache(config);
    }

    @Override
    public void checkHealthyBath(Collection<Serializable> serverIds) {
        serverIds.forEach(id -> checkHealthy(id));
    }

    @Override
    public void concurrentHeartBeatBath(Collection<Serializable> ids) {
        if (ids == null || ids.isEmpty()) {
            log.info("Heart beat failed,because 'ids' is null,but it will run next time.");
            return;
        }
        int pageSize = 3;
        int pageTotal = (int) Math.ceil(ids.size() / (pageSize * 1.0));
        for (int i = 0; i < pageTotal; i++) {
            Collection<Serializable> collect = ids.stream().skip(i * pageSize).limit(pageSize).collect(Collectors.toCollection(ArrayList::new));
            CompletableFuture<Void> future = CompletableFuture.runAsync(() -> checkHealthyBath(collect));
        }
    }

    //    public void concurrentHeartBeat(Collection<Serializable> ids) {
//        ids.forEach(id -> CompletableFuture.runAsync(() -> checkHealthy(id)));
//    }

    private static ReentrantReadWriteLock heartBeatWorkSerial = new ReentrantReadWriteLock();

    //定时 job 入口 3000ms
    @Override
    public void heartBeatWork() {
        ReentrantReadWriteLock.WriteLock writeLock = heartBeatWorkSerial.writeLock();
        if (!writeLock.tryLock()) {
            log.info("Heart beat server busy,run next time");
            return;
        }

        try {
            Collection<Serializable> serverIds = configProvider.getKeys();
            concurrentHeartBeatBath(serverIds);
//            log.info("Heart beat successfully");
        } catch (Exception e) {
            log.error("Heart beat failed cause by :{}", e.getMessage());
        } finally {
            writeLock.unlock();
        }

    }

}
