package com.gct.dlview.thrift;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.StampedLock;
import java.util.stream.Collectors;

@Component
@Slf4j
public class MemoryConfigProvider implements ConfigProvider, InitializingBean, DisposableBean {

    @Autowired
    private PersistProvider persistProvider;

    private HashMap<Serializable, ServerConfig> memory;

    private static StampedLock memoryLock = new StampedLock();

    private ConcurrentLinkedQueue<ServerConfig> memorySync = new ConcurrentLinkedQueue<>();

    private long readLock() {
        long stampRead = 0L;
        try {
            stampRead = memoryLock.tryReadLock(100, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            stampRead = 0;
            log.warn("Can not get lock from memory to read cause by:{}", e.getMessage());
        }
        if (stampRead == 0L) log.warn("Can not get lock from memory to read cause by: access timeout");
        return stampRead;
    }

    @Override
    public Collection<Serializable> getKeys() {
        long l = readLock();
        if (l == 0) return null;
        Collection<Serializable> collect = memory.keySet().stream().collect(Collectors.toCollection(ArrayList::new));
        memoryLock.unlockRead(l);
        return collect;
    }

    @Override
    public ServerConfig getValue(Serializable id) {
        long l = readLock();
        if (l == 0) return null;

        ServerConfig config = memory.get(id);
        ServerConfig re = Objects.isNull(config) ? null : config.clone();
        memoryLock.unlockRead(l);
        return re;
    }

    @Override
    public void updateCache(ServerConfig config) {
        try {
            memorySync.add(config);
        } catch (Exception e) {
            log.warn("update queue failed,confirm size:{}", memorySync.size());
        }
    }

    //定时job 入口  刷盘\库 1200ms
    @Override
    public void flush() {
        long writeLock = memoryLock.writeLock();
        try {
            HashMap<Serializable, ServerConfig> persistedConfig = persistProvider.getAllConfig();
            if (persistedConfig == null || persistedConfig.size() == 0)
                log.info("Persistent data read error，data is null");
            else {
                //delete from memory
                Set<Serializable> temp = memory.values().stream().map(ServerConfig::getId).filter(id -> persistedConfig.get(id) == null).collect(Collectors.toSet());
                temp.forEach(id -> memory.remove(id));

                persistedConfig.values().forEach(con -> {
                    ServerConfig config = memory.get(con.getId());
                    if (config == null) memory.put(con.getId(), con);//add to memory
                    else con.setRunState(config.getRunState());//update to persist
                });
                persistProvider.updateRunStateBatch(persistedConfig);
            }

        } catch (Exception ignore) {
            log.info("persist config error");
        } finally {
            memoryLock.unlockWrite(writeLock);
        }

    }

    //定时 job 同步缓存 100ms
    @Override
    public void syncMemory() {
        if (memorySync.size() == 0) return;
        long writeLock = memoryLock.tryWriteLock();
        if (writeLock == 0L) return;

        int i;//一轮50个，防止卡手
        for (i = 1; i <= 50; i++) {
            ServerConfig one = memorySync.poll();
            if (Objects.isNull(one)) break;
            memory.put(one.getId(), one);
        }
//        log.info("sync config ,from queue to memory master,completed size:{}", i - 1);
        memoryLock.unlockWrite(writeLock);
    }

    @Override
    public void afterPropertiesSet() throws Exception {

        memory = persistProvider.getAllConfig();
        if (Objects.isNull(memory) || memory.isEmpty())
            log.warn("Not found available server config");
    }

    @Override
    public void destroy() throws Exception {
        flush();
        memory.clear();
        log.info("Persist config successfully when destroy");
    }
}
