package com.gct.dlview.thrift;

import com.gct.dlview.config.ApplicationEventPublisher;
import com.gct.dlview.entity.TrainTask;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.StampedLock;
import java.util.stream.Collectors;

@Component
@Slf4j
public class MemoryTrainTaskProvider implements TrainTaskProvider, InitializingBean, DisposableBean {

    @Autowired
    PersistProvider persistProvider;

    private HashMap<Serializable, DLTrainTask> memory;

    private static final StampedLock memoryLock = new StampedLock();

    private final ConcurrentLinkedQueue<DLTrainTask> memorySync = new ConcurrentLinkedQueue<>();

    private long readLock() {
        long stampRead = 0L;
        try {
            stampRead = memoryLock.tryReadLock(100, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            stampRead = 0;
            log.warn("Can not get lock from memory to read cause by:{}", e.getMessage());
        }
        if (stampRead == 0L) log.warn("Can not get lock from memory to read cause by: access timeout");
        return stampRead;
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        HashMap<Serializable, TrainTask> allTrainTask = persistProvider.getAllTrainTask();
        if (Objects.isNull(allTrainTask) || allTrainTask.isEmpty())
            log.warn("Not found available train task detail.");
        memory = new HashMap<>();
        allTrainTask.forEach((id, task) -> {
            memory.put(id, new DLTrainTask().accessFromTaskStorage(task));
        });

    }

    @Override
    public Collection<Serializable> getKeys() {
        long l = readLock();
        if (l == 0) return null;
        Collection<Serializable> collect = memory.keySet().stream().collect(Collectors.toCollection(ArrayList::new));
        memoryLock.unlockRead(l);
        return collect;
    }

    @Override
    public DLTrainTask getValue(Serializable id) {
        long l = readLock();
        if (l == 0) return null;

        DLTrainTask config = memory.get(id);
        DLTrainTask re = Objects.isNull(config) ? null : config.clone();
        memoryLock.unlockRead(l);
        return re;
    }

    @Override
    public void updateCache(DLTrainTask trainTask) {
        try {
            memorySync.add(trainTask);
        } catch (Exception e) {
            log.warn("update queue failed,confirm size:{}", memorySync.size());
        }
    }

    @Override
    public void flush() {
        long writeLock = memoryLock.writeLock();

        try {
            HashMap<Serializable, TrainTask> persistedTrainTask = persistProvider.getAllTrainTask();
            if (persistedTrainTask == null || persistedTrainTask.size() == 0)
                log.info("Persistent data read error，data is null");
            else {
                //delete from memory
                Set<Serializable> temp = memory.values().stream()
                        .map(dlTrainTask -> dlTrainTask.getTrainTaskCopy().getId())
                        .filter(id -> persistedTrainTask.get(id) == null).collect(Collectors.toSet());
                temp.forEach(id -> memory.remove(id));
                //update memory
                memory.forEach((id, task) -> task.accessFromTaskStorage(persistedTrainTask.get(id)));
                //add memory
                persistedTrainTask.forEach((id, task) -> memory.putIfAbsent(id, new DLTrainTask().accessFromTaskStorage(task)));

                //update storage
                List<TrainTask> trainTasks = memory.values().stream().map(DLTrainTask::saveToTaskStorage).collect(Collectors.toList());
                persistProvider.updateTrainTaskBatch(trainTasks);


            }
        } catch (Exception e) {
            log.info("persist train task error");
        } finally {
            memoryLock.unlockWrite(writeLock);
        }
    }

    @Override
    public void syncMemory() {
        if (memorySync.size() == 0) return;
        long writeLock = memoryLock.tryWriteLock();
        if (writeLock == 0L) return;

        int i;//一轮50个，防止卡手
        for (i = 1; i <= 50; i++) {
            DLTrainTask newOne = memorySync.poll();
            if (Objects.isNull(newOne)) break;
            DLTrainTask oldOne = memory.get(newOne.getTrainTaskCopy().getId());
            if (Objects.isNull(oldOne)) continue;//此任务已经被删除
            oldOne.accessFromSyncQueue(newOne);

            memory.put(newOne.getTrainTaskCopy().getId(), oldOne);
        }
        log.info("sync train task ,from queue to memory master,completed size:{}", i - 1);
        memoryLock.unlockWrite(writeLock);
    }

    @Override
    public void destroy() throws Exception {
        flush();
        memory.clear();
        log.info("Persist train task successfully when destroy");
    }
}
