package com.gct.dlview.thrift;

import com.alibaba.fastjson.JSON;
import com.gct.dlview.entity.TrainTask;
import com.gct.dlview.service.GroupSampleRelationService;
import com.gct.dlview.service.SampleDiagramInterpretationTempService;
import com.gct.dlview.thrift.train.TrainParam;
import com.gct.dlview.thrift.train.TrainServerWrapper;
import com.gct.dlview.thrift.train.TrainStepController;
import io.netty.util.internal.ObjectUtil;
import io.swagger.models.auth.In;
import lombok.extern.slf4j.Slf4j;
import org.apache.thrift.TException;
import org.apache.thrift.protocol.TBinaryProtocol;
import org.apache.thrift.transport.TSocket;
import org.apache.thrift.transport.TTransport;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationListener;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;
import rx.Completable;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.stream.Collectors;

import static com.gct.dlview.entity.TaskStateEnum.AUTO_RETRY;
import static com.gct.dlview.entity.TaskStateEnum.INTERRUPT;

@Component
@Slf4j
public class TrainTaskManager implements TrainTaskExecute {

    @Autowired
    TrainTaskProvider trainTaskProvider;

    @Autowired
    ConfigProvider configProvider;

    @Autowired
    private TaskExecuteStep<DLTrainTask, TrainServerWrapper>[] trainTaskSteps;

    @Autowired
    TrainStepController stepController;


    @Override
    public boolean pushTask(DLTrainTask task) {
        if (Objects.isNull(task)) return false;
        if (!assignServer(task)) return false;

        try (TrainServerWrapper server = new TrainServerWrapper(configProvider.getValue(task.getTrainTaskCopy().getTrainServerId()))) {
            int step = stepController.checkControlSignal(task, server);
            if (step >= 0)
                try (TaskExecuteStep<DLTrainTask, TrainServerWrapper> run = trainTaskSteps[step].setServer(server)) {
                    //执行失败只能是发生了已知内部异常，中断任务，根据日志，维修bug后重新执行
                    if (!run.execute(task)) {
                        log.info("Execute step {},task {}", step, task.getTrainTaskCopy().getId());
//                        task.getTrainTaskCopy().setTaskState(AUTO_RETRY.getId());
                    }
                } catch (Exception ignore) {
//                    task.getTrainTaskCopy().setTaskState(INTERRUPT.getId());
                    //这里肯定是接口运行错误，或者负载不够，
                }
            return true;//返回true表示java端此处运行没有bug,所有影响均应同步
        } catch (Exception ignore) {

            return false;//java端产生了未知错误，则上面的操作都将不会产生影响
        }


    }

    @Override
    public boolean assignServer(DLTrainTask task) {
        TrainTask actualTask = task.getTrainTaskCopy();
        //只有初始状态的任务才需要分配，也就是说非初始状态的任务暂不支持切换负载的服务
        //TODO:重启后可以重新计算负载？
        if (actualTask.getTaskState() != 0) return true;

        if (Objects.nonNull(actualTask.getTrainServerId()) && actualTask.getTrainServerId() > 0 && actualTask.getEnableAutoLoadBalance() != 1)
            return true;

        boolean completed = false;
        //先指定，后自动负载到一个可用服务实例
        if (Objects.nonNull(actualTask.getTrainServerId()) && actualTask.getTrainServerId() > 0) {
            ServerConfig config = configProvider.getValue(actualTask.getTrainServerId());
            if (Objects.isNull(config)) return false;//等待下一轮次负载

            try (TrainServerWrapper server = new TrainServerWrapper(config)) {
                completed = server.canExchangeTrainTask(actualTask.getId().toString()) == 0;
            } catch (Exception ignore) {
            }

        }
        //没有指定或者指定的服务不可用，开启自动负载
        if (!completed && actualTask.getEnableAutoLoadBalance() == 1) {
            Collection<Serializable> keys = configProvider.getKeys();
            if (Objects.isNull(keys) || keys.isEmpty()) return false;//等待下一轮次负载
            Set<Serializable> confirm = new HashSet<>();
            for (Serializable key : keys) {
                ServerConfig value = configProvider.getValue(key);
                if (Objects.nonNull(value)) try (TrainServerWrapper server = new TrainServerWrapper(value)) {
                    confirm.add(key);
                    completed = server.canExchangeTrainTask(actualTask.getId().toString()) == 0;
                    if (completed) {
                        actualTask.setTrainServerId((Integer) key);
                        break;
                    }
                } catch (Exception ignore) {
                }
            }
            if (!completed && confirm.size() != keys.size()) return false;//发生了竞争锁，下次重新负载

            //若以上没有负载成功，使用负载策略
            if (!completed) {
                log.info("Train servers are busy,not assign for task->id:{}，then use random strategy", actualTask.getTrainServerId());
                int keyIndex = ThreadLocalRandom.current().nextInt(0, keys.size());
                actualTask.setTrainServerId((Integer) new ArrayList<>(keys).get(keyIndex));
                completed = true;
            }

        }

        return completed;
    }

    @Override
    public void pullTaskStat(String key, List<String> params, List<String> logs) {
        /**
         * 拉取的方式，判断是否需要拉取的任务复杂度较高，需要判断是否终止拉取，恢复拉取，拉取有效与否等等
         *
         * 改用c端推送的方式
         */
        if (StringUtils.isEmpty(key)) return;
        if (Objects.isNull(params) || params.isEmpty())
            if (Objects.isNull(logs) || logs.isEmpty())
                return;

        DLTrainTask task = trainTaskProvider.getValue(Integer.valueOf(key));
        if (Objects.isNull(task)) {//竞争锁获取不到的时候 retry whatever
            try {
                Thread.sleep(200);
            } catch (InterruptedException ignore) {
            }

            pullTaskStat(key, params, logs);
        }

        boolean updated = false;
        try {
            task.getTrainTaskCopy().setIntermediateData(TrainParam.trans(params).json());
            updated = true;
        } catch (Exception ignore) {

        }
        try {
            List<String> mLogs = JSON.parseObject(task.getTrainTaskCopy().getExecuteLog(), List.class);
            for (String log : logs) {
                mLogs.add(log);
                if (mLogs.size() > 15) mLogs.remove(mLogs.size() - 15);
            }
            task.getTrainTaskCopy().setExecuteLog(JSON.toJSONString(mLogs));
            updated = true;
        } catch (Exception ignore) {

        }
        if (updated) trainTaskProvider.updateCache(task);
    }

    /**
     * 将此方法注入到 thrift server 端
     *
     * @param key
     * @param params
     * @param logs
     */
    @Override
    public void asyncPullTaskStat(String key, List<String> params, List<String> logs) {
        CompletableFuture<Void> future = CompletableFuture.runAsync(() -> pullTaskStat(key, params, logs));
    }


    //region 分配可用训练服务 弃用
    @Deprecated
    private static final ReentrantReadWriteLock workSerial = new ReentrantReadWriteLock();
    @Deprecated
    private static final Set<Integer> assignedKeys = new HashSet<>();

    /**
     * 分配可用训练服务
     */
    @Deprecated
    public void workAssign() throws InterruptedException {
        if (!workSerial.writeLock().tryLock(100, TimeUnit.MILLISECONDS)) return;//串行分配

        try {
            for (Serializable taskId : trainTaskProvider.getKeys()) {
                if (assignedKeys.contains((Integer) taskId)) continue;
                DLTrainTask trainTask = trainTaskProvider.getValue(taskId);
                if (assignServer(trainTask)) {
                    assignedKeys.add((Integer) taskId);
                    trainTaskProvider.updateCache(trainTask);
                }
            }
        } catch (Exception e) {
            log.error("push task state failed cause by :{}", e.getMessage());
        } finally {
            workSerial.writeLock().unlock();
        }
    }
    //endregion


    /**
     * cache 中的推送一个task更新事件
     *
     * @param event cache 中的推送一个task更新事件,用以触发运行下一步
     */
    @Override
    public void onApplicationEvent(PushTrainTaskEvent event) {
        DLTrainTask task = (DLTrainTask) event.getSource();
        if (pushTask(task))
            trainTaskProvider.updateCache(task);
    }
}
