package com.gct.dlview.thrift.train;

import com.alibaba.fastjson.JSONObject;
import com.gct.dlview.thrift.ServerConfig;
import com.gct.dlview.thrift.ServerRunState;
import com.gct.dlview.thrift.TrainTaskExecute;
import lombok.extern.slf4j.Slf4j;
import org.apache.thrift.TException;
import org.apache.thrift.protocol.TBinaryProtocol;
import org.apache.thrift.transport.TSocket;
import org.apache.thrift.transport.TTransport;
import org.apache.thrift.transport.TTransportException;
import org.springframework.util.StringUtils;

import java.io.Closeable;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;


/**
 * create by lloyd at 2021-11-05 11:22
 * <p>
 * train server 的包装类，关注于连接、非法值处理等
 * <p>
 *
 * <p>
 * 特性：线程隔离，安全，缓存使用高，慎用
 * </p>
 */
@Slf4j
public class TrainServerWrapper implements TrainServer.Iface, Closeable {

    private static final ThreadLocal<TrainServer.Client> client = new ThreadLocal<>();
    private static final ThreadLocal<TTransport> transport = new ThreadLocal<>();
    private static final ThreadLocal<ServerConfig> config = new ThreadLocal<>();
    /**
     * flag for client if available
     */
    private final ThreadLocal<Boolean> available=new ThreadLocal<>();

    public TrainServerWrapper(ServerConfig config) {
        TrainServerWrapper.config.set(config);
    }

    /**
     * @return connected return true or return false
     */
    private boolean connect() {
        if (available.get()) return true;//avoid repeat initialize the connection

        close();

        ServerConfig serverConfig = TrainServerWrapper.config.get();
        if (Objects.isNull(serverConfig) || StringUtils.isEmpty(serverConfig.getHost()) || StringUtils.isEmpty(serverConfig.getPort()) || StringUtils.isEmpty(serverConfig.getTimout())) {
            log.error("The config.get() config is invalid.");
            available.set(false);
            return false;
        }

        transport.set(new TSocket(serverConfig.getHost(), serverConfig.getPort(), serverConfig.getTimout()));
        client.set(new TrainServer.Client(new TBinaryProtocol(transport.get())));
        try {
            transport.get().open();
            available.set(true);
            return true;
        } catch (TTransportException ignore) {
            log.warn("Can not connect the server by config:{}", JSONObject.toJSONString(serverConfig));
            available.set(false);
            return false;
        }
    }

    @Override
    public int healthyCheck() {
        if (!connect()) return ServerRunState.MISSING.getCode();
        try {
            return client.get().healthyCheck();
        } catch (TException e) {
            available.set(false);
            return ServerRunState.MISSING.getCode();
        }
    }

    /**
     * @param trainKey It's necessary.
     * @param localPath 自己已经封装无需传入
     * @return It's a secret.
     */
    @Override
    public boolean accessApply(String trainKey, String localPath) {
        if (!connect()) return false;
        try {
            return client.get().accessApply(trainKey, config.get().getLocalOriginPath());
        } catch (TException e) {
            available.set(false);
            return false;
        }
    }

    /**
     * @param trainKey 获取任何数据时需要双向验证所属关系
     * @return -1为预期外数据
     */
    @Override
    public int checkTrainState(String trainKey) {
        if (!connect()) return -1;
        try {
            return client.get().checkTrainState(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return -1;
        }
    }

    @Override
    public boolean isSafetyPoint(String trainKey) {
        if (!connect()) return false;
        try {
            return client.get().isSafetyPoint(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public int executeStepControl(String trainKey, int nextStep) {
        if (!connect()) return TrainStepState.DEFAULT.getId();
        try {
            return client.get().executeStepControl(trainKey, nextStep);
        } catch (TException ignore) {
            available.set(false);
            return TrainStepState.DEFAULT.getId();
        }
    }

    /**
     * 日志获取采取追加方式。不要直接替换全部，具体持久化多少由上层调用者控制
     *
     * @param trainKey 获取任何数据时需要双向验证所属关系
     * @return empty || list
     */
    @Override
    @Deprecated
    public List<String> getLogText(String trainKey) {
        if (!connect()) return new ArrayList<>();
        try {
            return client.get().getLogText(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return new ArrayList<>();
        }
    }

    /**
     * 运行参数，缓存级别
     * 如果没有一是因为服务处于空闲期，再就是因为服务验证失败，还有就是因为所属关系不对应
     *
     * @param trainKey 获取任何数据时需要双向验证所属关系
     * @return empty || string
     */
    @Override
    @Deprecated
    public List<String> getProcessParam(String trainKey) {
        if (!connect()) return new ArrayList<>();
        try {
            return client.get().getProcessParam(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return new ArrayList<>();
        }
    }

    //region train log server interface

    /**
     * spring inject
     */
    private TrainTaskExecute execute;

    public TrainServerWrapper(TrainTaskExecute execute) {
        this.execute = execute;
    }

    /**
     * 此为服务端接口，java应实现该接口不应使用
     *
     * @param trainKey unique constant key for zhe train task
     * @param params task param ,auto wrap or unwrap
     * @param logs logs.
     * @return It's a secret.
     */
    @Override
    public boolean pullTrainStat(String trainKey, List<String> params, List<String> logs) {
        execute.asyncPullTaskStat(trainKey, params, logs);
        return true;
    }
    //endregion

    /**
     * #是否可以切换任务
     * 当值大于0的时候应按照约定处理完成即可
     *
     * @param newTrainKey 预期加入的新任务key
     * @return 0：可以 1：日志未同步完成 2：参数未同步完成 3：不是安全点请等待 4：原有任务未过最大存活期不予切换
     */
    @Override
    public int canExchangeTrainTask(String newTrainKey) {
        if (!connect()) return -1;
        try {
            return client.get().canExchangeTrainTask(newTrainKey);
        } catch (TException ignore) {
            available.set(false);
            return -1;
        }
    }

    @Override
    public boolean removeTrainTask(String trainKey) {
        if (!connect()) return false;
        try {
            return client.get().removeTrainTask(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean setProcessParam(String trainKey, List<String> params) {
        if (!connect()) return false;
        try {
            return client.get().setProcessParam(trainKey, params);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    @Deprecated
    public boolean setSampleCount(int num) {
        if (!connect()) return false;
        try {
            return client.get().setSampleCount(num);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean transSGTSampleOne(String trainKey, List<List<Double>> sample, int tag) {
        if (!connect()) return false;
        try {
            return client.get().transSGTSampleOne(trainKey, sample, tag);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean transSGTSampleBatch(String trainKey, List<List<List<Double>>> samples, List<Integer> tags) {
        if (!connect()) return false;
        try {
            return client.get().transSGTSampleBatch(trainKey, samples, tags);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean transSGTSamplePage(String trainKey, List<List<List<Double>>> samples, List<Integer> tags, int curPage, int pageSize, int pageTotal) {
        if (!connect()) return false;
        try {
            return client.get().transSGTSamplePage(trainKey, samples, tags, curPage, pageSize, pageTotal);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean executeStandard(String trainKey) {
        if (!connect()) return false;
        try {
            return client.get().executeStandard(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean executeHOGAnalysis(String trainKey) {
        if (!connect()) return false;
        try {
            return client.get().executeHOGAnalysis(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public boolean executeTrain(String trainKey) {
        if (!connect()) return false;
        try {
            return client.get().executeTrain(trainKey);
        } catch (TException ignore) {
            available.set(false);
            return false;
        }
    }

    @Override
    public void close() {
        try {
            if (Objects.isNull(transport.get()) && transport.get().isOpen()) transport.get().close();
        } catch (Exception ignore) {
        }
        client.remove();
        transport.remove();
    }
}
